/*
    file: LsdSrc/lsd.c
    Copyright(C)2000 CNRS-Laboratoire de Pharmacognosie-UMR 6013-Jean-Marc Nuzillard

    This file is part of LSD.

    LSD is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    LSD is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with LSD; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "defs.h"
#include <stdio.h>
#include <string.h>

extern struct noy nucl[MAX_NUCL] ; /* lecture.c */
extern void	lecture(void) ; /* lecture.c */
extern void	start(void) ; /* start.c */
extern int	estlie(int x, int y) ; /* start.c */
extern void	about(void) ; /* start.c */
extern void	wrln2(int y) ; /* start1.c */
extern void	display(int p) ; /* display.c */
extern int	doublelia(void) ; /* double.c */
extern int	antibredt(void) ; /* double.c */
extern int	nssl ; /* sub.c */
extern int	init_sstr(void) ; /* sub.c */
extern int	attribuer(void) ; /* sub.c */
extern void	initequ(void) ; /* equiv.c */
extern int	calcstat(struct atom *a) ; /* equiv.c */
extern int	testequ(int x) ; /* equiv.c */
extern void	modequ(int x) ; /* equiv.c */
extern void	marque(void) ; /* heap.c */
extern void	empile(int t, int a, int v) ; /* heap.c */
extern void	relache(void) ; /* heap.c */
extern void	histoire(void) ; /* heap.c */
extern void	remcorr3(int fl, int n, int x, int y) ; /*remcorr.c */
extern void	inutil1(int x, int y) ; /*remcorr.c */
extern void	remcorr2(int x, int z) ; /*remcorr.c */
extern void	inutil2(int x, int z, int y) ; /*remcorr.c */
extern void	inutil4(int z, int x, int y) ; /*remcorr.c */
extern void	remcorrvar(int x, int z) ; /*remcorr.c */

void	ouvrir(int argc, char *argv[]) ;
void	work(void) ;
void	chlies(void) ;
int	insatur(int x) ;
void	queue(int x, int y) ;
void	j1hyp(int x, int y) ;
void	j2hyp(int x, int y) ;
void	j2_1(int x, int y) ;
void	j2_2(int x, int y) ;
int	useprop(int x, int y) ;
int	usep(int x, int y) ;
int	valide(int x, int y) ;
int	goodsp2(int x, int flx) ;
void	ajbase(int x) ;
void	wrbase(void) ;
void	paires(void) ;
void	pair1(void) ;
void	finir(void) ;
int	dupsol(void) ;
void	classe(int n, int t[]) ;
int	ok(void) ;
void	piege(void) ;

FILE *fic1, *fic2 ; /* input, output file */

struct	atom at[MAX_ATOMES] ; /* all the informations about atoms */
int	list[MAX_LIST][MAX_ATOMES] ; /* logical lists */
struct	pile heap[MAX_HEAP] ; /* heap for the optimization of recursivity */
int	valcorr[MAX_CORR] ; /* to know if correlations have been used */
int	max_atomes = 0 ; /* highest atom number */
int	icorr = 0 ; /* correlations counter */
int	phase = 1 ; /* 1, 2 or 3, see work */
int	nivheap = 0 ; /* heap level */
int	flpaires = FALSE ; /* for final atom pairing */
int	iessai ; /* for final atom pairing */
int	icomp ; /* for final atom pairing */
int	comp[MAX_ATOMES] ; /* for final atom pairing */
int	essai[MAX_BOND][2] ; /* for final atom pairing */
int	isol = 0 ; /* solution number */
int	iph2 = 0 ; /* number of entries into the pairing process */
int	flabout = FALSE ; /* to display input final interpretation */
int	flhistoire = FALSE ; /* to display the resolution path */
int	fldisp = 0 ; /* 0 : displays soutions as bond, 1: for JMNMOL */
int	flverb = FALSE ; /* verbosity, 0, 1, 2 */
int	flpart = FALSE ; /* gives partial solutions */
int	flbackph2 ; /* true upon coming back successfully from atoms pairing */
int	flstep = FALSE ; /* step by step operation */
int	flwork = TRUE ; /* true if the solution(s) has to be found */
int	fldupl = TRUE ; /* true if duplicated solutions must be removed */
int	flsubs = TRUE ; /* true if substructural info must be considered */
int	base[MAX_ATOMES] ; /* determines the order of the use of correlations */
int	nbase = 0 ; /* number of atoms in the base */
ucar	solclas[MAX_ATOMES][2] ; /* compressed solution */
ucar	bufsol[MAX_BUFSOL][2] ; /* solutions storage */
int	lnsize[MAX_LN] ; /* sizes of numerical lists */
int	ln[MAX_LN][MAX_LNBUF] ; /* numerical lists */
int	flhmbvar = FALSE ; /* true if hmbc correlation(s) have variants */
int	supniv = 0 ; /* highest heap level */
int	nivdisp = 0 ; /* heap level at which full program stop is required */

main(int argc, char *argv[])
/*
 * input, output file set-up and resolution top-level
 */
{
	/* header */

	fprintf(stderr, "LSD-3.0\n") ;
	fprintf(stderr, "Copyright(C)2000 CNRS - UMR 6013 - Jean-Marc Nuzillard.\n") ;
	fprintf(stderr, "Please read the LICENSE file.\n") ;
	fprintf(stderr, "jm.nuzillard@univ-reims.fr.\n") ;

	ouvrir(argc, argv) ;    /* defines input, output files */
	lecture() ;/* reads input file */
	start() ;  /* input file analysis */
	nssl = init_sstr() ;    
/* immediately after calling start(), nssl is the true number of subbonds.
 * after init_sstr it contains the number of recursive steps in the
 * assignent process */
	heap[0].niv = 0 ;  /* heap initialisation */
	initequ() ;/* equivalence classes initialisation */
	if (flabout) {
		about() ;
	}
/* displays the starting state of the resolution process */
	if (flwork) {
		work() ;   
/* solution(s) search */
		if ((fldisp == 1) && (isol)) {
			fprintf(fic2, "0\n") ;
/* means EOF for OUTLSD */
		}
	}
	if (flwork && (isol == 0) && (nivdisp == 0)) {
/* no solution has been found. Prints the highest reached heap level (nivdisp)
 * rerunning lsd with the "MLEV nivdisp" command will dump the resolution
 * state at its most advanced point */
		fprintf(stderr, "max stack level : %d\n", supniv) ;
	}
	if (isol) {
		fprintf(stderr, "%d solution%s found.\n", 
			isol, (isol>1) ? "s" : "") ; 
	}
        exit(0) ;
} /* main */


void	ouvrir(int argc, char *argv[])
/*
 * ouvrir means "to open". sets FILE* variables according to lsd call line 
 */
{
	char	nom1[40], nom2[40] ;

	switch (argc) {
	case 1 :
/* lsd without argument : uses standard devices */
		fic1 = stdin ;
		fic2 = stdout ;
		break ;
	case 2 :
/* lsd with one argument. the file "file" is opened to read from,
 * the file "file.sol" is opened to write into */
		strcpy(nom1, argv[1]) ;
		strcpy(nom2, argv[1]) ;
		fic1 = fopen(nom1, "r");
		if (fic1 == NULL) {
			printf("Cannot open %s\n", nom1) ;
			exit(2) ;
		}
		strcat(nom2, ".sol");   /* output file */
		fic2 = fopen(nom2, "w");
		if (fic2 == NULL) {
			printf("Cannot create %s\n", nom2) ;
			exit(3) ;
		}
		break ;
	default :
		printf("Usage : lsd [file]\n") ;
		exit(1) ;
		break ;
	}
} /* ouvrir */


void	work(void)
/*
 * Warning : serious stuff begins here !
 * the variable phase controls the resolution process. It has been initialised
 * to 1. Work is called recursively from chlies() (means : make bonds from
 * C-H correlations) and from paires()
 */
{
	switch (phase) {
	case 1 : 
		chlies() ; 
/* uses correlations */
		break ;
	case 2 : 
		paires() ; 
/* uncomplete atoms pairing */
		break ;
	case 3 : 
		finir() ; 
/* validity checks and printout */
		break ; 
	}
} /* work */

void	chlies(void) 
/*	
 * phase 1 : uses correlations
 */
{
	int	go, i, x, y, n ;

	go = TRUE ;	
/* will be false when a suitable correlating atom x will be found */
	if (flverb > 1) {
		wrbase() ;
	}
/* the base is an initially empty set that is progressively filled  
 * with the atoms most recently involved in bonds. If the base is not empty
 * correlations of atoms present in the base are considered in priority.
 * When all the correlations of an atom from the base are used this
 * atom is removed from the base. */
	if (nbase) {
/* the base is not empty */
		for (i = 0 ; (i < 5) && go ; i++) {
			for (x = 1 ; (x <= max_atomes) && (go = !( base[x] && (insatur(x) == i))) ; x++) 
				;
/* atoms from the base have all at least one usable correlation.
 * x has i free positions */
		}
	} else {
/* the base is empty */
		for (i = 0 ; (i < 5) && go ; i++) {
			for (x = 1 ; (x <= max_atomes) && (go = !( at[x].utile && (insatur(x) == i) && (at[x].nc != 0))) ; x++)
				;
/* x has a usable correlation and i free positions */
		}
	}
	if (i == 5) {
/* x has not been found. All the correlation data have been exploited */
		iph2++;	
/* counts the number of times the program has entered into the phase #2
 * this has been useful for debugging */
		marque() ;
		empile(0, (int)&phase, 2) ;
/* now phase equals 2 */
		work() ;
/* recursive call */
		relache() ;
/* back to the original state */
	} else {	
/* x is found. now a correlation of x with y has to be found */
		for (i = 0 ; (n = at[x].ex[i], !valcorr[n]) ; i++) 
			;	
/* the first one which is valid will be the good one. This is x's ith 
 * correlation which is also the nth correlation */
		marque() ;
		empile(0, (int)&valcorr[n], FALSE) ; 
/* it is not valid anymore */
		y = at[x].other[i] ;	
/* x correlates with y */
		if (flverb) {	/* message */
			printf("%d : analysis of correlation %d - ",  nivheap, x) ;
			wrln2(y) ;
		}
		remcorr3(0, n, x, y) ;	
/* correlation number n between x and y is removed. the first argument (0)
 * indicates that this removal is due to the choice of a correlation */
		empile(1, x, y) ;	
/* the choice of this correlation will be mention in the history listing */
		queue(x, y) ; 	
/* analyses the possible consequences of the x-y correlation */
		relache() ;
	}
}


int	insatur(int x)	
/*
 * returns x's numbur of free positions
 */
{
	return at[x].nlmax - at[x].nlia ;
}


void	queue(int x, int y0) /* exploitation de la correlation x-y */
{
	int	i, y ;

	if (y0 > 0) {
/* x correlates with an atom, not with a group of atomes */
		j1hyp(x, y0) ; 
/* considers first they can be birectly bound */
		j2hyp(x, y0) ;
/* consideres then they can be bound to a common atom */
	} else {
		for (i = 0 ; i < lnsize[-y0] ; i++) {
			y = ln[-y0][i] ;
/* y correlates possibly with x */
			if (testequ(y)) {
/* y has to the available according to equivalence status */
				marque() ;
				modequ(y) ;
/* updates y's equivalence class (if any) */
				empile(5, x, y) ;
/* the choice of y as a substitute for y0 is registered for the
 * history listing */
				if (flverb) {
					printf("%d : hypothesis : correlation %d - %d\n", nivheap, x, y) ;
				}
				j1hyp(x, y) ;
				j2hyp(x, y) ;
/* x and y are possibly either bound together or bound to a common atom. */
				relache() ;
			}
		}
	}
}


void	j1hyp(int x, int y)
/*
 * binds x and y, performs housekeeping
 */
{
	if (insatur(x) && insatur(y)) {
/* both x and y have to own more then 0 free position */
		marque() ;
		if (flverb > 1) {
			printf("%d : tries j1 %d - %d\n", nivheap, x, y) ;
		}
		if (useprop(x, y) && valide(x, y)) {
/* the x-y bond has to be compatible with x's and y's property information.
 * housekeeping for x and y */
			if (flverb) {
				printf("%d : *** OK for j1 %d - %d\n", nivheap, x, y) ;
			}
			if (ok()) { 
/* eventual user's agreement */
				ajbase(x) ;
				ajbase(y) ;
/* base management */
				empile(2, x, y) ;
				inutil1(x, y) ;
				inutil1(y, x) ;
/* removes eventual confirmation correlations arising from the new x-y bond */
				work() ;
/* goes on */
			}
		}
		relache() ;
	}
}


void	j2hyp(int x, int y)
/*
 * in j2_1 and j2_2 the program will look for z to be bound to x and y.
 * if x is complete then z is obviously one of x's neighbours. This
 * situation is handled by j2_1(x, y). x and y roles are symmetrical.
 */
{
	if (insatur(x)) {
		if (insatur(y)) {
			j2_2(x, y) ;
		} else {
			j2_1(y, x) ;
		}
	} else {
		if (insatur(y)) {
			j2_1(x, y) ;
		} else {
                        return ;
/*
 * this cannot happen when x and y are bound to a common atom because
 * x and y would be visible, and because correlations between visible
 * atoms are systematically removed before this point is reached. 
 * therefore x and y must have non-overlapping neighbours sets, and
 * this is in contradiction with the existence if a correlation between them.
 * The program must backtrack if this point is reached. Subtle isn't it?
 */
		}
	}
}


void	j2_1(int x, int y)
/*
 * finds the intermediate between x and y as one of x's neighbours when
 * x is already complete.
 */
{
	int	i, z ;

	for (i = 0 ; i < at[x].nlia ; i++) {
		z = at[x].lia[i] ;
		if (insatur(z) && testequ(z)) {
/* z is x's ith neighbour, it can at least be bound to a new atom
 * and complies to equivalence constraints */ 
			marque() ;
			if (flverb > 1) {
				printf("%d : tries j2 %d - %d - %d\n", nivheap, x, z, y) ;
			}
			if (useprop(y, z) && valide(y, z)) {
/* the y-z bond is compatible with y and z properties. uptades data about
 * x and y */
				if (flverb) {
					printf("%d : *** OK for j2 %d - %d - %d\n", nivheap, x, z, y) ;
				}
				if (ok()) {
/* upon eventual user's agreement */
					modequ(z) ;
/* useless ? z was already bound to x */
					ajbase(y) ;
					ajbase(z) ;
/* updates the content of the base */
					empile(2, y, z) ;
/* the y-z bond is stored for history listing */
					inutil2(y, z, x) ;
/* removes eventual confirmation correlations */
					work() ;
/* goes on */
				}
			}
			relache() ;
		}
	}
}


void	j2_2(int x, int y)
/* 
 * finds the intermediate z between x and y when neither x nor y is complete
 */
{
	int	z, flx, fly, m, valx, valy ;

	for (z = 1 ; z <= max_atomes ; z++) {
		if (at[z].utile && (z != x) && (z != y) && testequ(z)) {
/* z is different from x and from y and complies to equivelence contraints */
			flx = estlie(x, z) ;
/* if true there is no need to bind x and z */
			fly = estlie(y, z) ;
/* if true there is no need to bind y and z */
			if (flx && fly) {
/* x and y would be visible. it is impossible because the pending
 * correlation shoud have been removed previously */
				printf("%3d et %3d ne devraient pas correler\n", x, y) ;
				histoire() ;
				about() ;
				exit(1) ;
/* never happened to me. */
			}
			m = (flx || fly) ? 1 : 2 ;
/* number of bonds to provide to z */
			if (insatur(z) >= m) {
/* z has enough free positions */
				marque() ;
				if (flverb > 1) {
					printf("%d : tries j2 %d - %d - %d\n", nivheap, x, z, y) ;
				}
				valx = (flx) ? TRUE : (useprop(x, z) && valide(x, z)) ;
/* checks for the x-z bond */
				valy = valx &&  ((fly) ? TRUE : (useprop(y, z) && valide(y, z))) ;
/* if OK checks for z-y bond */
				if (valy) {
/* everything is OK, proceeds to housekeeping */
					if (flverb) {
						printf("%d : *** OK for j2 %d - %d - %d\n", nivheap, x, z, y) ;
					}
					if (ok()) {
						if (!flx) {
/* binds x to z */
							modequ(z) ;
							ajbase(x) ;
							ajbase(z) ;
							empile(2, x, z) ;
							inutil2(x, z, y) ;
						}
						if (!fly) {
/* binds y to z */
							modequ(z) ;
							ajbase(y) ;
							ajbase(z) ;
							empile(2, y, z) ;
							inutil2(y, z, x) ;
						}
						work() ;
/* goes on */
					}
				}
				relache() ;
			}
		}
	}
}


int	useprop(int x, int y)
/*
 * returns true if x and y can be neighbours according to their properties
 */
{
	return usep(x, y) && usep(y, x) ;
}


int	usep(int x, int y)
/*
 * return true if y can be bound to x according to x's properties 
 */
{
	int	i, n, *a ;

	n = at[x].nlpr ;
/* n is the number of x's property lists */
	if (!n) {
		return TRUE ;
/* no property */
	}
	for (i = 0 ; (i < n) && (!list[at[x].lpr[i]][y]) ; i++) 
		;
/* y is in x's ith property list */
	if (i == n) {
		a = &at[x].extocc ;
	} else {
		a = &at[x].occ[i] ;
	}
/* a : address of the number of occurences to be decremented */
	if (!(*a)) {
		return FALSE ;
/* the number of occurences cannot become negative */
	}
	empile(0, (int)a, (*a) - 1) ;
/* decremented */
	return TRUE ;
}


int	valide(int x, int y)
/*
 * bond formation housekeeping
 */
{
	int	n, flx, fly, *a, nm ;

	a = &at[x].nlia ;
/* address of the actual number of atoms bound to x */
	n = (*a) + 1 ;
/* what this number will be */
	nm = at[x].nlmax ;
/* its maximum value */
	if (n > nm) {
		return FALSE ;
/* of course */
	}
	flx = (n == nm) ;
/* x becomes complete */
	empile(0, (int)a, n) ;
/* updates the actual number of atoms bound to x */
	at[x].lia[n-1] = y ;
/* y is the last x's neighbour */

	a = &at[y].nlia ;
	n = (*a) + 1 ;
	nm = at[y].nlmax ;
	if (n > nm) 
		return FALSE ;
	fly = (n == nm) ;
	empile(0, (int)a, n) ;
	at[y].lia[n-1] = x ;
/* as well for y */

	return goodsp2(x, flx) && goodsp2(y, fly) ;
/* a complete sp2 atom must have at least one sp2 neighbour */
}


int	goodsp2(int x, int flx)
/*
 * a complete sp2 atom must have exactly one sp2 neighbour
 */
{
	int	i, s ;

	if ((!flx) || (at[x].hybrid == 0)) {
		return TRUE ;
/* no point if either x is not complete or x is sp3 */
	}
	s = 0 ;
	for (i = 0 ; i < at[x].nlmax ; i++) {
		s += at[at[x].lia[i]].hybrid ;
	}
/* s is the number of x's sp2 neighbours */
	return s ;
}


void	ajbase(int x)
/*
 * attempts to add x into the base 
 */
{
	if (!base[x] && at[x].nc) {
/* x must not be already in the base and must have at least one valid 
 * correlation */
		empile(0, (int)&base[x], TRUE) ;
		empile(0, (int)&nbase, nbase + 1) ;
	}
}


void	wrbase(void)
/*
 * writes the base's content
 */
{
	int	x ;

	if (!nbase)  {
		printf("empty connected set\n") ;
	} else {
		printf("\t*%d : ", nbase);
		for (x = 1 ; x <= max_atomes ; x++) {
			if (base[x]) {
				printf("%d ", x) ;
			}
		}
		printf("\n") ;
	}
}


void	paires(void)
{
	int	x ;

	if (flpaires) {
/* when pairing initialisation has been performed */
		pair1() ;
	}
	else {
/* pairing initialisation */
		icomp = 0 ;
		for (x = 1 ; x <= max_atomes ; x++) {
			if (at[x].utile && insatur(x)) {
				comp[icomp++] = x ;
			}
		}
/* the table comp contains the icomp uncomplete atoms */
		marque() ;
		flbackph2 = FALSE ;
/* will become true only in partial mode and if pairing and 
 * phase #3 are successfull */
		empile(0, (int)&flpaires, TRUE) ;
		iessai = 0 ;
/* will be the number of times pair1 will be called recursively. it happens
 * each time a new bond is established, and the bound atoms will
 * be stored in essai[1][iessai] and essai[2][iessai] */
		work() ;
/* goes on */
		if (flbackph2 && !dupsol()) {
			isol++;
			display(fldisp) ;
/* in partial mode the solution is displayed here because only bonds
 * established during phase #1 are present */
		}
		relache() ;
	}
}


void	pair1(void)
{
	int	i, j, k, x, y, stoppe ;

	for (i = 0 ; (i < icomp) && (x = comp[i], !(insatur(x) && testequ(x))) ; i++)
		;
/* x the first available uncomplete atom from comp, at index i */
	if (i == icomp) {
/* no such x, goes to phase #3 */
		marque() ;
		empile(0, (int)&phase, 3) ;
		work() ;
		relache() ;
	} else {
/* ok for x */
		marque() ;
		modequ(x) ;
		for (j = i + 1 ; j < icomp ; j++) {
			y = comp[j] ;
/* y will be tentatively bound to x */
			if (x == y) {
				continue ;
/* how could it be possible ? */
			}
			if (!insatur(y)) {
				continue ;
/* y must still be uncomplete */
			}
			if (!testequ(y)) {
				continue ;
/* looks for equivalence */
			}
			if (estlie(x, y)) {
				continue ;
/* cannot bind x and y if it is already done */
			}
			if (flverb > 1) {
				printf("%d : tries to bind %d - %d\n", nivheap, x, y) ;
			}
			stoppe = FALSE ;
			for (k = iessai ; k > 0 &&  (essai[k][1] == x) &&  (stoppe = (essai[k][2] > y)) ; k--) 
				;
/* goes back through the previously established bonds where x is the
 * starting atom. If any, the previous y must not be greater than the actual 
 * one */
			if (stoppe) continue ;
			marque() ;
			if (useprop(x, y) && valide(x, y)) {
/* bonds housekeeping */
				if (flverb) 
					printf("%d : *** OK to bind %d - %d\n", nivheap, x, y) ;
				if (ok()) {
					modequ(y) ;
/* looks for equivalence */
					empile(0, (int)&iessai, iessai + 1) ;
					essai[iessai][1] = x ;
					essai[iessai][2] = y ;
					empile(4, x, y) ;
/* stores x and y for the history listing */
					work() ;
/* attempts no establish one more bond */
				}
			}
			relache() ;
			if (flbackph2) {
/* in partial mode, if phase #3 has been successfully there is no need
 * to look for other pairing possibilities */
				break ;
			}
		}
		relache() ;
	}
}


void	finir(void)
/*
 * final tests before solution delivery 
 */
{
	if (!doublelia()) {
		return ;
/* impossible to set double bonds : backtrack */
	}
	if (antibredt()) {
		return ;
/* antibredt solution : backtrack */
	}
	if (!attribuer()) {
		return ;
/* impossible to find the substructure : backtrack */
	}
	if (flpart) {
		flbackph2 = TRUE ;
/* the solution is a good one. in partial mode the test for unicity has
 * to be performed within the partial structures set */
		return ;
	} else if (dupsol()) {
/* only original solutions are printed */
		return ;
	}
	isol++;
	display(fldisp) ;
/* hurray !!!! */
	if (flhistoire) {
		histoire() ;
	}
}


int	dupsol(void)
/*
 * return true when the solution is duplicated
 * should always return false when flhmbvar is false
 * this idea is worth being tested...
 * dupsol stores the actual solution if it is an original one
 */
{
	int	icl, i, j, x, y, clas[4], is, is1, t ;

	if(!fldupl) {
		return FALSE ;
	}
/* filter for duplicated solutions may be disabled */
	icl = 0 ;
	for (x = 1 ; x <= max_atomes ; x++) {
		if (at[x].utile) {
			for (i = 0, j = 0 ; i < at[x].nlia ; i++)
				if (x < (y = at[x].lia[i])) 
					clas[j++] = y ;
			classe(j, clas) ;
/* x's neighbours have been ranked in ascending order */
			for (i = 0 ; i < j ; i++, icl++) {
				solclas[icl][0] = (ucar)x ;
				solclas[icl][1] = (ucar)clas[i] ;
			}
		} /* solution en ordre */
	}
/* solclass contains the actual solution stored as a set of characters 
 * now it will be compared to the previous solutions */
	is = 0 ;
	for (i = 0 ; i < isol ; i++) {
		t = bufsol[is][0] ;
/* t is the number of bonds within the ith solution */
		if (t == icl) {
			is1 = is + 1 ;
/* is1 is positionned at the beginning of the ith solution in bufsol */
			for (j = 0 ; (j < t) && (solclas[j][0] == bufsol[is1][0]) && (solclas[j][1] == bufsol[is1][1]) ; j++,is1++) 
				;
			if (j == t) {
				return TRUE ;
/* the solution is already known */
			}
		}
		is += t + 1 ;
/* for the next solution in bufsol */
	} 
/* the solution is a new one */
	bufsol[is][0] = icl ;
/* stores the number of bonds within the actual solution */
	is++;
	for (i = 0 ; is < MAX_BUFSOL && i < icl ; i++, is++) {
		bufsol[is][0] = solclas[i][0] ;
		bufsol[is][1] = solclas[i][1] ;
	}
/* stores the bonds */
	if (is == MAX_BUFSOL) {
/* solutions storage is full */
		printf("modify MAX_BUFSOL\n") ;
		exit(1) ;
	}
	return FALSE ;
} /* the new solution is stored */


void	classe(int n, int t[])
/*
 * the most stupid sorting algorithm in the world 
 */
{
	int	i, j, x ;

	for (i = 0 ; i < n - 1 ; i++) 
		for (j = i + 1 ; j < n ; j++)
			if (t[i] > t[j]) {
				x = t[i] ;
				t[i] = t[j] ;
				t[j] = x ;
			}
}


int	ok(void)
/*
 * when flstep is true, ask the user to validate choices
 */
{
	if (!flstep) 
		return TRUE ;
	printf("<CR>, n(o), f(inish), q(uit) ? ") ;
	switch (getchar()) {
	case 'n' : 
		getchar() ; 
		return FALSE ; 
	case 'f' : 
		getchar() ; 
		flstep = FALSE ; 
		return TRUE ; 
	case 'q' : 
		getchar() ; 
		exit(0) ; 
		break ;
	default : 
		return TRUE ; 
	}
}


void	piege(void)
/*
 * used for debugging with edge
 */
{
	printf("****************** coucou\n") ;
}


